/
How do I compress a file?
A) Use the -19! operator. e.g.
-19!(`:sourceFile;`:targetFile;logicalBlockSize;compressionAlgorithm;compressionLevel)
where
logicalBlockSize is a power of 2 between 12 and 20 
(pageSize or allocation granularity to 1MB - pageSize for AMD64 is 4kB, 
sparc is 8kB. Windows seems to have a default allocation granularity of 64kB). 
Note that this argument affects both compression speed and compression ratio 
(larger blocks can be slower and better compressed)
compressionAlgorithm is one of the following, 0 - none, 1 - kdb+ ipc, 2 - gzip
compressionLevel is between 0 and 9 
(valid only for gzip, use 0 for other algorithms)
\

n:1000000

`:db/test1 set asc n?100; /create a test data file

/compress input file test, to output file ztest, 
/using a block size of 128kB (2 xexp 17) and kdb+ IPC compression algorithm
show -19!(`:db/test1;`:db/test1_ipc;17;1;0)
 
/using a block size of 128kB (2 xexp 17) and gzip compression algorithm
/note: on windows, need to install the dll called zlibwapi.dll in order to use gzip option
show -19!(`:db/test1;`:db/test1_gzip;17;2;6) 
 /
/ check the compressed data is the same as the uncompressed data
(get[`:db/test1]~get`:db/test1_ipc)and(get[`:db/test1]~get`:db/test1_gzip)
/ 
/Set has been over-writen in such a way that we can create both single-file
/and splayed tables on disk in compressed format
/For example:
/Single file
(`:db/test2_ipc;17;1;0) set asc n?`3 /use kdb+ IPC compression algorithm
(`:db/test2_gzip;17;2;6) set asc n?`3 /use gzip compression algorithm 
 
/splayed
t1:([]sym:asc n?`3;time:.z.p+til n;price:n?1000.;size:n?100);
(`:db/zsplay1/;17;1;0) set .Q.en[`:db;t1] /use kdb+ IPC compression algorithm
(`:db/zsplay2/;17;2;6) set .Q.en[`:db;t1] /use gzip compression algorithm 

show"Both splayed tables are loaded back into memory as identical tables"; 
(get `:db/zsplay1)~(get `:db/zsplay2) 
 
/kdb+ compressed files/splays can also be appended to 
(`:db/zippedTest;17;1;0) set 100000?10; /create sinlge binary file in compressed format
`:db/zippedTest upsert 100000?10; /append more data to the compressed file
-21!`:zippedTest
 
 
/
EOD function .u.end on RDB with compression enabled:
/old function (no compression)
.u.end:{[x]
	t:tables`.; /get a lost of all tables within the default namespace
	t@:where `g=attr each t@\:`sym; /get a list of tables with g attribute on sym
	.Q.hdpf[`$":",.u.x 1;`:.;x;`sym];/save tables to disk in date partitioned format
	@[;`sym;`g#] each t; /reapply g attribute to sym column of tables
	};



/new function (with compression)
/
inputs:
tbl_name - table to save to disk
hdb_directory - location on disk of hdb
dt - date partition we are writing data to
pcol - sort table on this column and apply p attribute
\
save_date_partitioned_with_compression:{[tbl_name;hdb_directory;dt;pcol]
	tbl:xasc[pcol;value tbl_name]; /create copy of table sorted in ascending order on pcol (likely sym)
	tbl:@[tbl;pcol;`p#]; /Apply parted attribute to pcol
	/save table to disk in compressed, date partitioned format
	set[(` sv (hdb_directory;`$string dt;tbl_name;`);17;1;0);
		.Q.en[hdb_directory;tbl]];
	delete from tbl_name; /clear out table
 };

.u.end:{[x] 
	t:tables`.; /get a lost of all tables within the default namespace
	t@:where `g=attr each t@\:`sym; /get a list of tables with g attribute on sym
	save_date_partitioned_with_compression[;`:.;x;`sym] each t;
	@[;`sym;`g#] each t; /reapply g attribute to sym column of tables
	(neg hopen hsym `$hdl_hdb)"\\l ." /tell HDB to refresh
	};
 
 
 
 
 
 
 /
 